@@ -21,13 +21,18 @@
 '''A Python library for reading and writing C3D files.'''
 
 import array
-import numpy
-import struct
-import operator
 import cStringIO
+import numpy as np
+import operator
+import struct
 import warnings
 
 
+PROCESSOR_INTEL = 84
+PROCESSOR_DEC = 85
+PROCESSOR_MIPS = 86
+
+
 class Header(object):
     '''Header information from a C3D file.
 
@@ -46,13 +51,13 @@ class Header(object):
     first_frame : int
         Index of the first frame of data.
 
-        This is actually not used in Phasespace data files; instead, to seek to
-        data, use the POINTS.ACTUAL_START_FIELD parameter.
+        This is actually not used in Phasespace data files; instead, the first
+        frame number is stored in the POINTS:ACTUAL_START_FIELD parameter.
     last_frame : int
         Index of the last frame of data.
 
-        This is actually not used in Phasespace data files; instead, to seek to
-        data, use the POINTS.ACTUAL_END_FIELD parameter.
+        This is actually not used in Phasespace data files; instead, the last
+        frame number is stored in the POINTS:ACTUAL_END_FIELD parameter.
     sample_per_frame : int
         Number of samples per frame. Seems to be unused in Phasespace files.
     frame_rate : float
@@ -71,8 +76,8 @@ class Header(object):
     def __init__(self, handle=None):
         '''Create a new Header object.
 
-        Arguments
-        ---------
+        Parameters
+        ----------
         handle : file handle, optional
             If given, initialize attributes for the Header from this file
             handle. The handle must be seek-able and readable. If `handle` is
@@ -103,8 +108,8 @@ class Header(object):
         This method writes exactly 512 bytes to the beginning of the given file
         handle.
 
-        Arguments
-        ---------
+        Parameters
+        ----------
         handle : file handle
             The given handle will be reset to 0 using `seek` and then 512 bytes
             will be written to describe the parameters in this Header. The
@@ -150,8 +155,8 @@ long_event_labels: {0.long_event_labels}
         This method reads exactly 512 bytes from the beginning of the given file
         handle.
 
-        Arguments
-        ---------
+        Parameters
+        ----------
         handle : file handle
             The given handle will be reset to 0 using `seek` and then 512 bytes
             will be read to initialize the attributes in this Header. The handle
@@ -159,8 +164,8 @@ long_event_labels: {0.long_event_labels}
 
         Raises
         ------
-        AssertionError, if the magic byte from the header is not 80 (the C3D
-        magic value).
+        AssertionError
+            If the magic byte from the header is not 80 (the C3D magic value).
         '''
         handle.seek(0)
         (self.parameter_block,
@@ -208,7 +213,7 @@ class Param(object):
                  desc='',
                  bytes_per_element=1,
                  dimensions=None,
-                 bytes=None,
+                 bytes='',
                  handle=None):
         '''Set up a new parameter with at least a name.
 
@@ -263,11 +268,19 @@ class Param(object):
             1 + len(self.desc) # size of desc and desc bytes
             )
 
-    def write(self, handle):
+    def write(self, group_id, handle):
         '''Write binary data for this parameter to a file handle.
 
-        This writes data at the current position in the file.
+        Parameters
+        ----------
+        group_id : int
+            The numerical ID of the group that holds this parameter.
+        handle : file handle
+            An open, writable, binary file handle.
         '''
+        handle.write(struct.pack('bb', len(self.name), group_id))
+        handle.write(self.name)
+        handle.write(struct.pack('<h', self.binary_size() - 2 - len(self.name)))
         handle.write(struct.pack('b', self.bytes_per_element))
         handle.write(struct.pack('B', len(self.dimensions)))
         handle.write(struct.pack('B' * len(self.dimensions), *self.dimensions))
@@ -285,37 +298,15 @@ class Param(object):
         self.bytes_per_element, = struct.unpack('b', handle.read(1))
         dims, = struct.unpack('B', handle.read(1))
         self.dimensions = [struct.unpack('B', handle.read(1))[0] for _ in range(dims)]
-        self.bytes = None
+        self.bytes = ''
         if self.total_bytes:
             self.bytes = handle.read(self.total_bytes)
         size, = struct.unpack('B', handle.read(1))
         self.desc = size and handle.read(size) or ''
 
-    def __str__(self):
-        '''Return a friendly string representation of this parameter.'''
-        kwargs = self.__dict__.copy()
-        kwargs['shaped'] = ''
-        if self.bytes:
-            kwargs['bytes'] = '{}{}'.format(
-                self.bytes[:10], ['', '...'][len(self.bytes) > 10])
-            if len(self.dimensions) == 2:
-                C, R = self.dimensions
-                kwargs['shaped'] = ' ->\n' + '\n'.join(
-                    repr(self.bytes[r * C:(r+1) * C]) for r in range(R))
-            if len(self.dimensions) == 0:
-                fmt = '<' + {2: 'H', 4: 'f'}.get(len(self.bytes), 'B')
-                kwargs['shaped'] = ' -> ' + struct.unpack(fmt, self.bytes)[0]
-
-        return '''\
-      name: {0.name}
-      desc: {0.desc}
- data_size: {0.data_size}
-dimensions: {0.dimensions}
-     bytes: {0.bytes:r}{0.shaped}'''.format(kwargs)
-
     def _as(self, fmt):
         '''Unpack the raw bytes of this param using the given struct format.'''
-        return struct.unpack(fmt, self.bytes)[0]
+        return struct.unpack('<' + fmt, self.bytes)[0]
 
     @property
     def int8_value(self):
@@ -357,6 +348,57 @@ dimensions: {0.dimensions}
         '''Get the param as a raw byte string.'''
         return self.bytes
 
+    def _as_array(self, fmt):
+        '''Unpack the raw bytes of this param using the given data format.'''
+        assert self.dimensions, \
+            '{}: cannot get value as {} array!'.format(self.name, fmt)
+        elems = array.array(fmt)
+        elems.fromstring(self.bytes)
+        return np.array(elems).reshape(self.dimensions)
+
+    @property
+    def int8_array(self):
+        '''Get the param as an array of 8-bit signed integers.'''
+        return self._as_array('b')
+
+    @property
+    def uint8_array(self):
+        '''Get the param as an array of 8-bit unsigned integers.'''
+        return self._as_array('B')
+
+    @property
+    def int16_array(self):
+        '''Get the param as a array of 16-bit signed integers.'''
+        return self._as_array('h')
+
+    @property
+    def uint16_array(self):
+        '''Get the param as a array of 16-bit unsigned integers.'''
+        return self._as_array('H')
+
+    @property
+    def int32_array(self):
+        '''Get the param as a array of 32-bit signed integers.'''
+        return self._as_array('i')
+
+    @property
+    def uint32_array(self):
+        '''Get the param as a array of 32-bit unsigned integers.'''
+        return self._as_array('I')
+
+    @property
+    def float_array(self):
+        '''Get the param as a array of 32-bit floats.'''
+        return self._as_array('f')
+
+    @property
+    def string_array(self):
+        '''Get the param as a array of raw byte strings.'''
+        assert len(self.dimensions) == 2, \
+            '{}: cannot get value as string array!'.format(self.name)
+        l, n = self.dimensions
+        return [self.bytes[i*l:(i+1)*l] for i in range(n)]
+
 
 class Group(dict):
     '''A group of parameters from a C3D file.
@@ -382,8 +424,8 @@ class Group(dict):
     def add_param(self, name, **kwargs):
         '''Add a parameter to this group.
 
-        Arguments
-        ---------
+        Parameters
+        ----------
         name : str
             Name of the parameter to add to this group. The name will
             automatically be case-normalized.
@@ -399,7 +441,25 @@ class Group(dict):
             1 + len(self.name) + # size of name and name bytes
             2 + # next offset marker
             1 + len(self.desc) + # size of desc and desc bytes
-            sum(p.binary_size() for p in self.params.itervalues()))
+            sum(p.binary_size() for p in self.itervalues()))
+
+    def write(self, group_id, handle):
+        '''Write this parameter group, with parameters, to a file handle.
+
+        Parameters
+        ----------
+        group_id : int
+            The numerical ID of the group.
+        handle : file handle
+            An open, writable, binary file handle.
+        '''
+        handle.write(struct.pack('bb', len(self.name), -group_id))
+        handle.write(self.name)
+        handle.write(struct.pack('<h', 3 + len(self.desc)))
+        handle.write(struct.pack('B', len(self.desc)))
+        handle.write(self.desc)
+        for param in self.itervalues():
+            param.write(group_id, handle)
 
     def get_int8(self, key):
         '''Get the value of the given parameter as an 8-bit signed integer.'''
@@ -451,11 +511,51 @@ class Manager(dict):
         '''Set up a new Manager with a Header.'''
         self.header = header or Header()
 
+    def check_metadata(self):
+        '''Ensure that the metadata in our file is self-consistent.'''
+        assert self.header.point_count == self.points_per_frame(), (
+            'inconsistent point count! {} header != {} parameter'.format(
+                self.header.point_count,
+                self.points_per_frame(),
+            ))
+
+        assert self.header.scale_factor == self.scale_factor(), (
+            'inconsistent scale factor! {} header != {} parameter'.format(
+                self.header.scale_factor,
+                self.scale_factor(),
+            ))
+
+        assert self.header.frame_rate == self.frame_rate(), (
+            'inconsistent frame rate! {} header != {} parameter'.format(
+                self.header.frame_rate,
+                self.frame_rate(),
+            ))
+
+        apf = self.analog_per_frame() * self.analog_frame_rate() / self.frame_rate()
+        assert self.header.analog_count == apf, (
+            'inconsistent analog count! {} header != {} analog per frame * '
+            '{} analog-fps / {} point-fps'.format(
+                self.header.analog_count,
+                self.analog_per_frame(),
+                self.analog_frame_rate(),
+                self.frame_rate(),
+            ))
+
+        start = self.get_uint16('POINT:DATA_START')
+        assert self.header.data_block == start, (
+            'inconsistent data block! {} header != {} parameter'.format(
+                self.header.data_block, start))
+
+        for name in ('POINT:LABELS', 'POINT:DESCRIPTIONS',
+                     'ANALOG:USED', 'ANALOG:LABELS', 'ANALOG:DESCRIPTIONS'):
+            if self.get(name) is None:
+                warnings.warn('missing parameter {}'.format(name))
+
     def add_group(self, group_id, name, desc):
         '''Add a new parameter group.
 
-        Arguments
-        ---------
+        Parameters
+        ----------
         group_id : int
             The numeric ID for a group to check or create.
         name : str, optional
@@ -465,11 +565,13 @@ class Manager(dict):
 
         Returns
         -------
-        A `Group` with the given ID, name, and description.
+        Group :
+            A :class:`Group` with the given ID, name, and description.
 
         Raises
         ------
-        KeyError, if a group with a duplicate ID or name already exists.
+        KeyError
+            If a group with a duplicate ID or name already exists.
         '''
         if group_id in self:
             raise KeyError(group_id)
@@ -482,8 +584,8 @@ class Manager(dict):
     def get(self, group, default=None):
         '''Get a group or parameter.
 
-        Arguments
-        ---------
+        Parameters
+        ----------
         group : str
             If this string contains a period (.), then the part before the
             period will be used to retrieve a group, and the part after the
@@ -495,8 +597,9 @@ class Manager(dict):
 
         Returns
         -------
-        Either a `Group` or a `Param` with the specified name(s). If neither is
-        found, returns the default value.
+        Group or Param :
+            Either a :class:`Group` or a :class:`Param` with the specified
+            name(s). If neither is found, returns the default value.
         '''
         try:
             return self[group]
@@ -506,23 +609,25 @@ class Manager(dict):
     def __getitem__(self, group):
         '''Get a group or parameter.
 
-        Arguments
-        ---------
+        Parameters
+        ----------
         group : str
-            If this string contains a period (.), then the part before the
-            period will be used to retrieve a group, and the part after the
-            period will be used to retrieve a parameter from that group. If this
-            string does not contain a period, then just a group will be
-            returned.
+            If this string contains a colon (:) or a period (.), then the part
+            before the punctuation will be used to retrieve a group, and the
+            part after the punctuation will be used to retrieve a parameter from
+            that group. If this string does not contain a period, then just a
+            group will be returned.
 
         Returns
         -------
-        Either a `Group` or a `Param` with the specified name(s).
+        Group or Param :
+            Either a :class:`Group` or a :class:`Param` with the specified
+            name(s).
 
         Raises
         ------
-        KeyError, if no group and parameter with the given identifiers are
-        found.
+        KeyError
+            If no group and parameter with the given identifiers are found.
         '''
         if isinstance(group, int):
             return super(Manager, self).__getitem__(group)
@@ -530,8 +635,10 @@ class Manager(dict):
         param = None
         if '.' in group:
             group, param = group.split('.', 1)
+        if ':' in group:
+            group, param = group.split(':', 1)
         group = super(Manager, self).__getitem__(group)
-        if param:
+        if param is not None:
             return group[param]
         return group
 
@@ -570,29 +677,43 @@ class Manager(dict):
     def parameter_blocks(self):
         '''Compute the size (in 512B blocks) of the parameter section.'''
         bytes = 4. + sum(g.binary_size() for g in self.itervalues())
-        return int(numpy.ceil(bytes / 512))
+        return int(np.ceil(bytes / 512))
 
     def frame_rate(self):
-        return self.header.frame_rate
+        return self.get_float('POINT:RATE')
 
     def scale_factor(self):
-        return self.get_float('POINT.SCALE')
+        return self.get_float('POINT:SCALE')
 
     def points_per_frame(self):
-        return self.get_uint16('POINT.USED')
+        return self.get_uint16('POINT:USED')
 
     num_points = points_per_frame
 
     def analog_per_frame(self):
-        return self.get_uint16('ANALOG.USED')
+        return self.get_uint16('ANALOG:USED')
 
     num_analog = analog_per_frame
 
-    def start_field(self):
-        return self.get_uint32('TRIAL.ACTUAL_START_FIELD')
+    def analog_frame_rate(self):
+        return self.get_float('ANALOG:RATE')
 
-    def end_field(self):
-        return self.get_uint32('TRIAL.ACTUAL_END_FIELD')
+    def point_labels(self):
+        return self.get('POINT:LABELS').string_array
+
+    def first_frame(self):
+        # this is a hack for phasespace files ... should put it in a subclass.
+        param = self.get('TRIAL:ACTUAL_START_FIELD')
+        if param is not None:
+            return param.int32_value
+        return self.header.first_frame
+
+    def last_frame(self):
+        # this is a hack for phasespace files ... should put it in a subclass.
+        param = self.get('TRIAL:ACTUAL_END_FIELD')
+        if param is not None:
+            return param.int32_value
+        return self.header.last_frame
 
 
 class Reader(Manager):
@@ -604,16 +725,15 @@ class Reader(Manager):
     construction:
 
     >>> r = c3d.Reader(open('capture.c3d', 'rb'))
-    >>> for points, analog in r.read_frames():
+    >>> for frame_no, points, analog in r.read_frames():
     ...     print('{0.shape} points in this frame'.format(points))
-    ...     frames.append(points, analog)
     '''
 
     def __init__(self, handle):
         '''Initialize this C3D file by reading header and parameter data.
 
-        Arguments
-        ---------
+        Parameters
+        ----------
         handle : file handle
             Read metadata and C3D motion frames from the given file handle. This
             handle is assumed to be `seek`-able and `read`-able. The handle must
@@ -622,8 +742,9 @@ class Reader(Manager):
 
         Raises
         ------
-        ValueError, if the processor metadata in the C3D file is anything other
-        than 84 (Intel format).
+        ValueError
+            If the processor metadata in the C3D file is anything other than 84
+            (Intel format).
         '''
         super(Reader, self).__init__(Header(handle))
 
@@ -633,70 +754,141 @@ class Reader(Manager):
         # metadata header
         buf = self._handle.read(4)
         _, _, parameter_blocks, processor = struct.unpack('BBBB', buf)
-        if processor != 84:
-            raise ValueError('We only read Intel C3D files.')
+        if processor != PROCESSOR_INTEL:
+            raise ValueError(
+                'we only read Intel C3D files (got processor {})'.
+                format(processor))
 
-        # read all metadata in a chunk, then process each chunk (to avoid block
-        # boundary issues).
-        bytes = self._handle.read(512 * parameter_blocks)
+        # read all parameter blocks as a single chunk to avoid block
+        # boundary issues.
+        bytes = self._handle.read(512 * parameter_blocks - 4)
         while bytes:
             buf = cStringIO.StringIO(bytes)
 
             chars_in_name, group_id = struct.unpack('bb', buf.read(2))
             if group_id == 0 or chars_in_name == 0:
+                # we've reached the end of the parameter section.
                 break
 
             name = buf.read(abs(chars_in_name)).upper()
+            offset_to_next, = struct.unpack('<h', buf.read(2))
 
-            offset_to_next, = struct.unpack('h', buf.read(2))
-
-            if group_id < 0:
+            if group_id > 0:
+                # we've just started reading a parameter. if its group doesn't
+                # exist, create a blank one. add the parameter to the group.
+                self.setdefault(group_id, Group()).add_param(name, handle=buf)
+            else:
+                # we've just started reading a group. if a group with the
+                # appropriate id exists already (because we've already created
+                # it for a parameter), just set the name of the group.
+                # otherwise, add a new group.
                 group_id = abs(group_id)
                 size, = struct.unpack('B', buf.read(1))
                 desc = size and buf.read(size) or ''
-                if self.get(name):
-                    warnings.warn('duplicate parameter group {}'.format(name))
+                group = self.get(group_id)
+                if group is not None:
+                    group.name = name
+                    group.desc = desc
+                    self[name] = group
                 else:
                     self.add_group(group_id, name, desc)
-            else:
-                self[group_id].add_param(name, handle=buf)
 
             bytes = bytes[2 + abs(chars_in_name) + offset_to_next:]
 
-    def read_frames(self):
+        self.check_metadata()
+
+    def read_frames(self, copy=True):
         '''Iterate over the data frames from our C3D file handle.
 
+        Parameters
+        ----------
+        copy : bool
+            If False, the reader returns a reference to the same data buffers
+            for every frame. The default is True, which causes the reader to
+            return a unique data buffer for each frame. Set this to False if you
+            consume frames as you iterate over them, or True if you store them
+            for later.
+
         Returns
         -------
-        This generates a sequence of (points, analog) ordered pairs, one
-        ordered pair per frame. The first element of each frame contains a numpy
-        array of 4D "points" and the second element of each frame contains a
-        numpy array of 1D "analog" values that were probably recorded
-        simultaneously.
+        sequence of (frame number, points, analog) :
+            This method generates a sequence of (frame number, points, analog)
+            tuples, one tuple per frame. The first element of each tuple is the
+            frame number. The second is a numpy array of parsed, 5D point data
+            and the third element of each tuple is a numpy array of analog
+            values that were recorded during the frame. (Often the analog data
+            are sampled at a higher frequency than the 3D point data, resulting
+            in multiple analog frames per frame of point data.)
+
+            The first three columns in the returned point data are the (x, y, z)
+            coordinates of the observed motion capture point. The fourth value
+            is an estimate of the error for this particular point, and the fifth
+            value is the number of cameras that observed the point in question.
+            Both the fourth and fifth values are -1 if the point is considered
+            to be invalid.
 
-        The four dimensions in the point data are typically (x, y, z) and a
-        "confidence" estimate for the point.
         '''
-        # find out where we seek to start reading frame data.
-        start_block = self.get_uint16('POINT.DATA_START')
-        if start_block != self.header.data_block:
-            if not start_block:
-                start_block = self.header.data_block
-
-        # read frame and analog data in either float or int format.
-        format = 'fi'[self.scale_factor() >= 0]
         ppf = self.points_per_frame()
         apf = self.analog_per_frame()
 
-        self._handle.seek((start_block - 1) * 512)
-        start = self._handle.tell()
-        f = 0
-        for f in xrange(self.end_field() - self.start_field() + 1):
-            points = array.array(format)
-            points.fromfile(self._handle, 4 * ppf)
-            analog = array.array(format)
-            analog.fromfile(self._handle, apf)
-            yield numpy.array(points).reshape((ppf, 4)), numpy.array(analog)
+        scale = abs(self.scale_factor())
+        is_float = self.scale_factor() < 0
+
+        point_dtype = [np.int16, np.float32][is_float]
+        point_scale = [scale, 1][is_float]
+        points = np.zeros((ppf, 5), float)
+
+        # TODO: handle ANALOG:BITS parameter here!
+        p = self.get('ANALOG:FORMAT')
+        analog_unsigned = p and p.string_value.strip().upper() == 'UNSIGNED'
+        analog_dtype = np.int16
+        if is_float:
+            analog_dtype = np.float32
+        elif analog_unsigned:
+            analog_dtype = np.uint16
+        analog = np.array([], float)
+
+        offsets = np.zeros((apf, ), int)
+        param = self.get('ANALOG:OFFSET')
+        if param is not None:
+            offsets = param.int16_array[:apf]
+
+        scales = np.ones((apf, ), float)
+        param = self.get('ANALOG:SCALE')
+        if param is not None:
+            scales = param.float_array[:apf]
+
+        gen_scale = 1.
+        param = self.get('ANALOG:GEN_SCALE')
+        if param is not None:
+            gen_scale = param.float_value
+
+        self._handle.seek((self.header.data_block - 1) * 512)
+        for frame_no in xrange(self.first_frame(), self.last_frame() + 1):
+            raw = np.fromfile(self._handle, dtype=point_dtype,
+                count=4 * self.header.point_count).reshape((ppf, 4))
+
+            points[:, :3] = raw[:, :3] * point_scale
+
+            valid = raw[:, 3] > -1
+            points[~valid, 3:5] = -1
+            c = raw[valid, 3].astype(np.uint16)
+
+            # fourth value is floating-point (scaled) error estimate
+            points[valid, 3] = (c & 0xff).astype(float) * scale
+
+            # fifth value is number of bits set in camera-observation byte
+            points[valid, 4] = sum((c & (1 << k)) >> k for k in range(8, 17))
+
+            if self.header.analog_count > 0:
+                raw = np.fromfile(self._handle, dtype=analog_dtype,
+                    count=self.header.analog_count).reshape((-1, apf))
+                analog = (raw.astype(float) - offsets) * scales * gen_scale
+
+            if copy:
+                yield frame_no, points.copy(), analog.copy()
+            else:
+                yield frame_no, points, analog
 
 
 class Writer(Manager):
@@ -711,8 +903,8 @@ class Writer(Manager):
     def __init__(self, handle):
         '''Initialize a new Writer with a file handle.
 
-        Arguments
-        ---------
+        Parameters
+        ----------
         handle : file handle
             Write metadata and C3D motion frames to the given file handle. This
             handle is assumed to be `seek`-able and `write`-able. The handle
@@ -730,6 +922,8 @@ class Writer(Manager):
 
     def write_metadata(self):
         '''Write metadata for this file to our file handle.'''
+        self.check_metadata()
+
         # header
         self.header.write(self._handle)
         self._pad_block()
@@ -739,34 +933,13 @@ class Writer(Manager):
         self._handle.write(struct.pack('BBBB', 0, 0, self.parameter_blocks(), 84))
         id_groups = sorted((i, g) for i, g in self.iteritems() if isinstance(i, int))
         for group_id, group in id_groups:
-            self._write_group(group_id, group)
+            group.write(group_id, self._handle)
 
         # padding
         self._pad_block()
         while self._handle.tell() != 512 * (self.header.data_block - 1):
             self._handle.write('\x00' * 512)
 
-    def _write_group(self, group_id, group):
-        '''Write a single parameter group, with parameters, to our file handle.
-
-        Arguments
-        ---------
-        group_id : int
-            The numerical ID of the group.
-        group : `Group`
-            The `Group` object to write to the handle.
-        '''
-        self._handle.write(struct.pack('bb', len(group.name), -group_id))
-        self._handle.write(group.name)
-        self._handle.write(struct.pack('h', 3 + len(group.desc)))
-        self._handle.write(struct.pack('B', len(group.desc)))
-        self._handle.write(group.desc)
-        for name, param in group.params.iteritems():
-            self._handle.write(struct.pack('bb', len(name), group_id))
-            self._handle.write(name)
-            self._handle.write(struct.pack('h', param.binary_size() - 2 - len(name)))
-            param.write(self._handle)
-
     def write_frames(self, frames):
         '''Write the given list of frame data to our file handle.
 
@@ -794,8 +967,8 @@ class Writer(Manager):
                               ):
         '''Write a set of frames to a file so it looks like Phasespace wrote it.
 
-        Arguments
-        ---------
+        Parameters
+        ----------
         frames : sequence of frame data
             The sequence of frames to write.
         frame_count : int
@@ -817,75 +990,68 @@ class Writer(Manager):
         # POINT group
         ppf = len(points)
         point_group = self.add_group(1, 'POINT', 'POINT group')
-        point_group.add_param('USED', desc='Number of 3d markers',
-                              data_size=2,
-                              bytes=struct.pack('H', ppf))
-        point_group.add_param('FRAMES', desc='frame count',
-                              data_size=2,
-                              bytes=struct.pack('H', min(65535, frame_count)))
-        point_group.add_param('DATA_START', desc='data block number',
-                              data_size=2,
-                              bytes=struct.pack('H', 0))
-        point_group.add_param('SCALE', desc='3d scale factor',
-                              data_size=4,
-                              bytes=struct.pack('f', point_scale_factor))
-        point_group.add_param('RATE', desc='3d data capture rate',
-                              data_size=4,
-                              bytes=struct.pack('f', point_frame_rate))
-        point_group.add_param('X_SCREEN', desc='X_SCREEN parameter',
-                              data_size=-1,
-                              dimensions=[2],
-                              bytes='+X')
-        point_group.add_param('Y_SCREEN', desc='Y_SCREEN parameter',
-                              data_size=-1,
-                              dimensions=[2],
-                              bytes='+Z')
-        point_group.add_param('UNITS', desc='3d data units',
-                              data_size=-1,
-                              dimensions=[len(point_units)],
-                              bytes=point_units)
-        point_group.add_param('LABELS', desc='labels',
-                              data_size=-1,
-                              dimensions=[5, ppf],
-                              bytes=''.join('M%03d ' % i for i in xrange(ppf)))
-        point_group.add_param('DESCRIPTIONS', desc='descriptions',
-                              data_size=-1,
-                              dimensions=[16, ppf],
-                              bytes=' ' * 16 * ppf)
+        point_group.add_param(
+            'USED', desc='Number of 3d markers', data_size=2,
+            bytes=struct.pack('<H', ppf))
+        point_group.add_param(
+            'FRAMES', desc='frame count', data_size=2,
+            bytes=struct.pack('<H', min(65535, frame_count)))
+        point_group.add_param(
+            'DATA_START', desc='data block number', data_size=2,
+            bytes=struct.pack('<H', 0))
+        point_group.add_param(
+            'SCALE', desc='3d scale factor', data_size=4,
+            bytes=struct.pack('<f', point_scale_factor))
+        point_group.add_param(
+            'RATE', desc='3d data capture rate', data_size=4,
+            bytes=struct.pack('<f', point_frame_rate))
+        point_group.add_param(
+            'X_SCREEN', desc='X_SCREEN parameter',
+            data_size=-1, dimensions=[2], bytes='+X')
+        point_group.add_param(
+            'Y_SCREEN', desc='Y_SCREEN parameter',
+            data_size=-1, dimensions=[2], bytes='+Z')
+        point_group.add_param(
+            'UNITS', desc='3d data units', data_size=-1,
+            dimensions=[len(point_units)], bytes=point_units)
+        point_group.add_param(
+            'LABELS', desc='labels', data_size=-1, dimensions=[5, ppf],
+            bytes=''.join('M%03d ' % i for i in xrange(ppf)))
+        point_group.add_param(
+            'DESCRIPTIONS', desc='descriptions', data_size=-1,
+            dimensions=[16, ppf], bytes=' ' * 16 * ppf)
 
         # ANALOG group
         apf = len(analog)
         analog_group = self.add_group(2, 'ANALOG', 'ANALOG group')
-        analog_group.add_param('USED', desc='analog channel count',
-                               data_size=2,
-                               bytes=struct.pack('H', apf))
-        analog_group.add_param('RATE', desc='analog frame rate',
-                               data_size=4,
-                               bytes=struct.pack('f', analog_frame_rate))
-        analog_group.add_param('GEN_SCALE', desc='analog general scale factor',
-                               data_size=4,
-                               bytes=struct.pack('f', gen_scale))
-        analog_group.add_param('SCALE', desc='analog channel scale factors',
-                               data_size=4,
-                               dimensions=[0])
-        analog_group.add_param('OFFSET', desc='analog channel offsets',
-                               data_size=2,
-                               dimensions=[0])
+        analog_group.add_param(
+            'USED', desc='analog channel count', data_size=2,
+            bytes=struct.pack('<H', apf))
+        analog_group.add_param(
+            'RATE', desc='analog frame rate', data_size=4,
+            bytes=struct.pack('<f', analog_frame_rate))
+        analog_group.add_param(
+            'GEN_SCALE', desc='analog general scale factor', data_size=4,
+            bytes=struct.pack('<f', gen_scale))
+        analog_group.add_param(
+            'SCALE', desc='analog channel scale factors', data_size=4,
+            dimensions=[0])
+        analog_group.add_param(
+            'OFFSET', desc='analog channel offsets', data_size=2,
+            dimensions=[0])
 
         # TRIAL group
         trial_group = self.add_group(3, 'TRIAL', 'TRIAL group')
-        trial_group.add_param('ACTUAL_START_FIELD', desc='actual start frame',
-                              data_size=2,
-                              dimensions=[2],
-                              bytes=struct.pack('I', 1))
-        trial_group.add_param('ACTUAL_END_FIELD', desc='actual end frame',
-                              data_size=2,
-                              dimensions=[2],
-                              bytes=struct.pack('I', frame_count))
+        trial_group.add_param(
+            'ACTUAL_START_FIELD', desc='actual start frame', data_size=2,
+            dimensions=[2], bytes=struct.pack('<I', 1))
+        trial_group.add_param(
+            'ACTUAL_END_FIELD', desc='actual end frame', data_size=2,
+            dimensions=[2], bytes=struct.pack('<I', frame_count))
 
         # sync parameter information to header.
         blocks = self.parameter_blocks()
-        point_group.params['DATA_START'].bytes = struct.pack('H', 2 + blocks)
+        point_group['DATA_START'].bytes = struct.pack('<H', 2 + blocks)
 
         self.header.data_block = 2 + blocks
         self.header.frame_rate = point_frame_rate
